/*
 * socks5 specific negotiation
 * marius aamodt eriksen (marius@umich.edu)
 */

#include "nylon.h"

#define SOCKS5_ATYP_IPV4 1
#define SOCKS5_ATYP_FQDN 3

#define SOCKS5_CD_CONNECT 1
#define SOCKS5_CD_BIND 2
#define SOCKS5_CD_UDP_ASSOC 3

/* request */
struct socks5_req {
	u_char vn;        /* version number */
	u_char cd;        /* command        */
	u_char rsv;       /* reserved       */
	u_char atyp;      /* address type   */
	u_int32_t d_addr; /* dest address   */
	u_int16_t d_p;    /* dest port      */
}; 

/* version reply */
struct socks5_v_repl {
	u_char ver;       /* version        */
	u_char res;       /* response       */
};

int socks5_connect __P((struct conf*, int, struct sockaddr_in*,
						struct socks5_req*));
int socks5_bind __P((struct conf*, int, struct sockaddr_in*,
					 struct socks5_req*));

int
negotiate5(struct conf *c, int client_s, int session_no)
{
	u_int i, j;
	int target_s;
	char *hostname;
	u_char num_methods, *methods, method, len; 
	struct sockaddr_in target_in;
	struct socks5_req req5;
	struct socks5_v_repl rep5;
	struct hostent *hent;	

	j = 1;
	i = method = 0;

	target_s = -ERROR_UNDEF;

	req5.vn = 5;
	req5.rsv = 0;
	
	printv(2, "Initiating SOCKS5 connection\n");

	/*
	 * start by retrieving number of methods, version number has
     * already been consumed by the calling procedure
	 */

	if (recvn(client_s, &num_methods, 1, 0) == -1) 
		return -ERROR_RECV;
	methods = (u_char*)malloc(num_methods);
	if (recvn(client_s, methods, num_methods, 0) == -1) {
		free(methods);
		return -ERROR_RECV;
	}
	free(methods);

	/*
	 * we don't support any authentication methods yet, so simply
	 * ignore it and send reply with no authentication required.
	 */

	rep5.ver = 5;
	rep5.res = 0;
	
	printv(2, "sending reply: version: %i method: %i ...", req5.vn, method);
	if (sendn(client_s, &rep5, 2, 0) == -1)
		return -ERROR_SEND;
	printv(2, "sent\n");

	/* receive up to atyp */
	if (recvn(client_s, &req5, 4, 0) == -1)
		return -ERROR_RECV;
	if (req5.vn != 5)
		return -ERROR_HEAD;

	bzero(&target_in, sizeof(target_in));

	switch (req5.atyp) {
	case SOCKS5_ATYP_IPV4:
		printv(2, "ipv4 address requested\n");
		if (recvn(client_s, &req5.d_addr, 4, 0) == -1) {
			printv(2, "Error receiving request");
			return -ERROR_RECV;
		}
		target_in.sin_family = AF_INET;
		target_in.sin_addr.s_addr = req5.d_addr;    /* already in
													 * network order */
		break;
	case SOCKS5_ATYP_FQDN:

		if (recvn(client_s, &len, 1, 0) == -1)
			return -ERROR_RECV;
		hostname = (char*)malloc(len+1);
		if (recvn(client_s, hostname, len, 0) == -1)
			return -ERROR_RECV;
		hostname[len] = '\0';
		hent = gethostbyname(hostname);
		target_in.sin_family = AF_INET;
		target_in.sin_addr = *(struct in_addr*)hent->h_addr;
		free(hostname);
		break;
	default:
		return -ERROR_ADDR;
	}

	if (recvn(client_s, &req5.d_p, 2, 0) == -1)
		return -ERROR_RECV;
	target_in.sin_port = req5.d_p;

	printv(1, "Target host:\n");
	printv(1, "\tIPv4 address : %s\n", inet_ntoa(target_in.sin_addr));
	printv(1, "\tPort         : %i\n", ntohs(target_in.sin_port));

	switch (req5.cd) {
	case SOCKS5_CD_CONNECT:
		target_s = socks5_connect(c, client_s, &target_in, &req5); 
		break;
	case SOCKS5_CD_BIND:
		/* target_s = socks5_bind(c, sfd, &target_addr, &req5); */
		break;
	case SOCKS5_CD_UDP_ASSOC:
		return -ERROR_UNSUP;
	default:
		return -ERROR_SOCKS5;
	}

	return target_s;
}

int
socks5_connect(struct conf *c, int client_s, struct sockaddr_in *target_in,
			   struct socks5_req *req5)
{
	int target_s, ret;
	struct sockaddr_in bind_in;
	
	if ((target_s = socket(AF_INET, SOCK_STREAM, 0)) < 0) 
		return -ERROR_SOCKET;

	printv(2, "making connection to target host...\n");
	
	if (c->if_conn) {
#ifdef WIN32
		/* XXX this needs to be fixed -- now using conf */
		struct hostent *cinfhost;
		cinfhost = gethostbyname(c->if_conf);
		bind_in.sin_addr = *(struct in_addr*)cinfhost->h_addr;
#else /* WIN32 */
		bind_in.sin_addr = get_ip_from_if(c->if_conn);
#endif /* WIN32 */
		bind_in.sin_port = 0;
		if (bind(target_s, (struct sockaddr*)&bind_in,
				 sizeof(bind_in)) < 0) {
			return -ERROR_BIND;
		}
	}
		
	if (connect(target_s,
				(struct sockaddr*)target_in, sizeof(struct sockaddr_in)) < 0) {
		printv(1, "unsuccessful connect!\n");
		req5->cd = 1;
	} else {
		printv(1, "successful connect!\n");
		req5->cd = 0;
	}
	
	printv(2, "made connection to target host...\n");
 	
	req5->atyp = SOCKS5_ATYP_IPV4; /* we are always sending an IPv4 addr back,
									* for now */ 
	
	if (sendn(client_s, req5, 10, 0) == -1) {
		ret = -ERROR_REPLY;
		goto cleanup;
	}

	if (req5->cd == 1) {
		ret = -ERROR_CONNECT;
		goto cleanup;
	}	
	
	/* success! */
	return target_s;
	
 cleanup:
	close(target_s);
	return ret;
}	

#if 0
int
socks5_bind(struct conf *c, int client_s, struct sockaddr_in *target_in,
			struct socks5_req *req5)
{
	int listen_sfd, accept_sfd;
	struct sockaddr_in serv_addr;
	
	memset(&serv_addr, 0, sizeof(serv_addr));

	if ((listen_sfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		/* XXX: fill out appropriate SOCKS5 error reply */
		printv(2, "error in creating socket\n");
		return -1;
	}
	
	if (bind(listen_sfd, (struct sockaddr*)addr,
			 sizeof(*addr))< 0) {
		printv(2, "error in binding socket\n");
		return -1;
	}
	
	if (listen(listen_sfd, 1) < 0) {
		printv(2, "error in setting listening status of socket\n");
		return -1;
	}

	/* send first reply */

	if (accept_sfd = accept(listen_sfd, (struct sockaddr*)&serv_addr,
							(socklen_t)sizeof(struct sockaddr))) {
		/* XXX - sizeof(struct sockaddr) may lie, check stevens */
		printv(2, "accept failed\n");
		return -1;
	}
	/* send second reply */
	/* give up socket to client */
}

#endif /* 0 */
